/*
 * mns-v.95beta
 * Multifuctional Network Scanner
 * bind & aempirei 11jan2k
 *
 * usage:
 * ./mns [options...] < <target> | <-f <file>> >
 *
 */

#include "include/mns.h"
#include <pcap.h>

extern int vuln(char *, int);

int usage(char *arg);
void load_defaults(int *);

u_short mns_opts = 0;
char *device = NULL;
char *scan_target = NULL;
char global[16] = { 0 };

void main (int argc, char *argv[]) {
   time_t total_t;
   struct utsname hinfo;
   char errbuf[PCAP_ERRBUF_SIZE];
   pcap_t *pd_temp;
   int children = 30;     /* number of processes to fork */
   int opt;               /* file descriptor for getopt() */
   int i, j;

   extern char *optarg;   /* optarg for getopt */
   extern int opterr;     /* opterr for getopt */

   setvbuf (stdout, NULL, _IONBF, 0);

   /* very cute centering technique */
   for(i = 0; i < 78; i++) putchar('*');
   putchar('\n');
   i = (78 - sizeof(VERSION " " AUTHORS)) >> 1;
   for(j = 0; j <= i; j++) putchar(' ');
   puts(VERSION " " AUTHORS);
   for(i = 0; i < 78; i++) putchar('*');
   putchar('\n');
   
   if (argc < 2) usage(argv[0]);

   load_defaults(&children);
   
   opterr = 0; /* we like pretty output */
  
   while ((opt = getopt(argc, argv, "F:C:I:RSVAufciqrwxv")) != -1) {
      switch (opt) {
         case 'F': /* limit child processes */
	    mns_opts |= HOSTLIST;
	    scan_target = optarg;
	    break;
	 case 'C': /* limit child processes */
	    children = atoi(optarg);
            if (children < 2 || children > 200) { /* validate children */
               puts(" * invalid number of children"
                    ", reverting to default");
               children = 30;
            }
	    break;
         case 'I': /* set active interface */
            if(!(device = malloc(strlen(optarg) + 1))) {
               printf(" * cannot process interface name\n");
               exit(EXIT_FAILURE);
            }
            strcpy(device, optarg);
            break;
         case 'R': /* resolve ips to hostnames */
	    mns_opts |= RESOLVE;
            break;
         case 'S': /* resolve ips to hostnames */
	    mns_opts |= SYNSCAN;
            break;
	 case 'V': /* display version information */
	    puts(VERSION);
	    exit(EXIT_FAILURE);
	    break;
	 case 'A': /* enable all scan checks */
	    mns_opts |= ALLVULNS;
	    break;
	 case 'u': /* qpop vuln scan */
	    mns_opts |= UDPSCAN;
	    break;
	 case 'q': /* qpop vuln scan */
	    mns_opts |= VULNQPOP;
	    break;
	 case 'i': /* imap vuln scan */
	    mns_opts |= VULNIMAP;
	    break;
	 case 'f': /* ftp vuln scan */
	    mns_opts |= VULNFTP;
	    break;
	 case 'c': /* cgi vuln scan */
	    mns_opts |= VULNCGI;
	    break;
	 case 'r': /* rpc vuln scan */
	    mns_opts |= VULNRPC;
	    break;
         case 'w': /* open wingate scan */
            mns_opts |= WINGATE;
            break;
         case 'x': /* open X servers */
            mns_opts |= XOPEN;
            break;
         case 'v': /* verbose output & promiscious scanning */
            mns_opts |= VERBOSE;
            break;
	 case '?': /* default case for unknown options */
	    usage(argv[0]);
	    break;
      }
   }

   if(optind < argc && !(mns_opts & HOSTLIST))
      scan_target = argv[optind];

   setuid (0);
   uname(&hinfo);
   
   printf("\nRunning on: '%s' running %s %s on a(n) %s\n\n",
      hinfo.nodename, hinfo.sysname, hinfo.release, hinfo.machine);

   if(getuid()) {
      printf(" * %s does not have root privilages\n", getlogin());
      exit(EXIT_FAILURE); /* get lost punk */
   }
   
   if(!scan_target) {
      puts(" * No target was specified");
      exit(EXIT_FAILURE);
   }
   
   if(!device) {
      if(!(device = pcap_lookupdev(errbuf))) {
         printf(" * pcap device lookup: %s\n", errbuf);
         exit(EXIT_FAILURE);
      }
   }

   if(!(pd_temp = pcap_open_live(device, 500, 0, 100, errbuf))) {
      printf(" * pcap device: %s\n", errbuf);
      exit(EXIT_FAILURE);
   }

   pcap_close(pd_temp);
  
   printf("Selected Interface: %s\n", device);

   if(mns_opts & ALLVULNS) {
      printf("Vulnerability checks: ");
      if(mns_opts & VULNQPOP) printf("qpop ");
      if(mns_opts & VULNIMAP) printf("imap ");
      if(mns_opts & VULNCGI) printf("cgi ");
      if(mns_opts & VULNRPC) printf("rpc ");
      if(mns_opts & VULNFTP) printf("ftp ");
      if(mns_opts & WINGATE) printf("wingate ");
      if(mns_opts & XOPEN) printf("xserv ");
      putchar('\n');
   }
   
   if(mns_opts & RESOLVE) printf("Resolving IPs to hostnames\n");
   if(mns_opts & SYNSCAN) printf("Forced SYN port scan mode\n");
   if(mns_opts & UDPSCAN) printf("UDP port scanning enabled\n");
   if(mns_opts & VERBOSE) printf("Verbose output mode selected\n");

   total_t = time(NULL);
   i = vuln(scan_target, children);
   total_t = time(NULL) - total_t;

   printf("\n-> %i machines scanned : ", i);
   printf("[%li min %li sec ] : ", total_t / 60, total_t % 60 );
   printf("%.02f seconds/machine <-\n\n", (float)total_t / i);
}

int usage (char *program) {
  printf(" usage: %s [options...] {target|-F file}\n", program);
  printf("   target           single scan where target is an IP Address or hostname\n"
         "   -F file          file containing a list of IP Addresses and hostnames\n"
	 "   [-qifcrwx]       enable vulnerability checks which consist of\n"
	 "                    the following scanning options\n"
	 "                       -q  QPOP\n"
	 "                       -i  IMAP\n"
	 "                       -f  FTP\n"
	 "                       -c  CGI\n"
	 "                       -r  RPC\n"
         "                       -w  WINGATE\n"
         "                       -x  XSERV\n"
         "   [-v]             verbose output and promiscious scanning\n"
         "   [-u]             scan for listening UDP service ports\n"
	 "   [-C children]    set maximum child processes (2-200)\n"
         "   [-I interface]   set active packet capturing interface\n"
	 "   [-V]             display version information\n"
	 "   [-R]             try to resolve IPs to hostnames\n"
	 "   [-A]             enable all vulerability checks\n"
         "   [-S]             force SYN port scanning\n\n");
   exit(EXIT_FAILURE);
}

void load_defaults(int *children) {
   FILE *f_in;
   char buffer[80];
   char *token;
   
   if(!(f_in = fopen(CONF_FILE, "r"))) {
      puts(" * error opening " CONF_FILE);
      exit(EXIT_FAILURE);
   }

   for(;;) {
      fgets(buffer, 80, f_in);
      if(feof(f_in)) break;

      if(!(token = strtok(buffer, " \t\n"))) continue;
      if(*token == '#') continue;
      if(!strcmp(token, "vulns")) {
         for(;;) {
            if(!(token = strtok(NULL, " \t\n"))) break;
            if(!strcmp(token, "all")) {
               mns_opts |= ALLVULNS;
               break;
            }
            if(!strcmp(token, "qpop")) mns_opts |= VULNQPOP;
            else if(!strcmp(token, "imap")) mns_opts |= VULNIMAP;
            else if(!strcmp(token, "ftp")) mns_opts |= VULNFTP;
            else if(!strcmp(token, "cgi")) mns_opts |= VULNCGI;
            else if(!strcmp(token, "rpc")) mns_opts |= VULNRPC;
            else if(!strcmp(token, "wingate")) mns_opts |= WINGATE;
            else if(!strcmp(token, "xxserv")) mns_opts |= XOPEN;
         }
      } else if(!strcmp(token, "verbose")) {
         if(!(token = strtok(NULL, " \t\n"))) continue;
         if(!strcmp(token, "on")) mns_opts |= VERBOSE;
      } else if(!strcmp(token, "children")) {
         if(!(token = strtok(NULL, " \t\n"))) continue;
         *children = atoi(token);
         if (*children < 2 || *children > 200) { /* validate children */
            puts(" * " CONF_FILE ": invalid number of children"
                           ", reverting to default");
            *children = 30;
         }
      } else if(!strcmp(token, "interface")) {
         if(!(token = strtok(NULL, " \t\n"))) continue;
         if(strcmp(token, "detect")) {
            if(!(device = malloc(strlen(token) + 1))) {
               printf(CONF_FILE " * cannot process interface name\n");
               fclose(f_in);
               exit(EXIT_FAILURE);
            }
            strcpy(device, token);
         }
      } else if(!strcmp(token, "resolve")) {
         if(!(token = strtok(NULL, " \t\n"))) continue;
         if(!strcmp(token, "on")) mns_opts |= RESOLVE;
      } else if(!strcmp(token, "synscan")) {
         if(!(token = strtok(NULL, " \t\n"))) continue;
         if(!strcmp(token, "on")) mns_opts |= SYNSCAN;
      } else if(!strcmp(token, "udpscan")) {
         if(!(token = strtok(NULL, " \t\n"))) continue;
         if(!strcmp(token, "on")) mns_opts |= UDPSCAN;
      }
   }
   fclose(f_in);
}
