#include "include/mns.h"
#include "include/inetraw.h"

static char *checkosmain(void);
void show_print(FILE *);

/* global varibles */
extern time_t timedelay;
struct fingerprint pak[PACKETS];

/* local structures */
struct fingerprint {
   u_short set:2,
           seq:2,
           ackset:3,
           urg:1;
   u_long ack;
   u_short win;
   u_short flag;
};

int checkos (struct in_addr target, FILE *fsp, int *plist,
                      struct in_addr hostaddress)
{
   char *buffer = NULL;
   int i, pcount;

   if (plist) pcount = plist[0] + 1;
   else return(-1);

   init_packet_devices(IPPROTO_TCP);
   
   local_ip = hostaddress;

   for (i = 1; i < min(pcount, 3) && !buffer; i++) {
      dst_port = htons(plist[i]);
      if((buffer = checkosmain())) {
         fprintf(fsp, "\n\tTCP Sequence Prediction Difficulty -> %c ",
                 *buffer);
	 switch (*buffer) {
            case '0': fprintf(fsp, "( indeterminate )"); break;
            case '1':
	    case '2': fprintf(fsp, "( easy )"         ); break;
	    case '3':
	    case '4': fprintf(fsp, "( moderate )"     ); break;
	    case '5': fprintf(fsp, "( hard )"         ); break;
	    case '6': fprintf(fsp, "( infeaseable )"  ); break;
	    case '7':
	    case '8': fprintf(fsp, "( impossible )"   ); break;
            default:  fprintf(fsp, "( hamming error )"); break;
         }
         fprintf(fsp, "\n\tOperating System -> %s", buffer + 1);
      }
   }

   if(!buffer) {
      fprintf(fsp, "\n\tOperating System -> ( indeterminate )");
      if(mns_opts & VERBOSE) show_print(fsp);
   }

   term_packet_devices();

   return(0);
}

static char *checkosmain(void) {
   FILE *fs_cfg;
   static char osname[60] = { 0 };
   struct tcphdr tcp;
   struct ip ip;
   hdrinfo s_ack;
   u_short start, s;
   u_long myseq[PACKETS];
   u_long seq_mask = 0, last_seq = 0;
   time_t timeout;
   char line[80];
   int n, i, ham;

   for (n = 0; n < PACKETS; n++)
      bzero(&pak[n], sizeof(struct fingerprint));

   srand(time (NULL) & 0x0000ffff);
   start = s = (rand() % 5000) + 5000;
   s_ack.ack = 0;

#define sendtcppacket(a, b)\
   do{\
      s_ack.seq = myseq[a] = rand();\
      src_port = htons(s + a);\
      sendtcp(&s_ack, b);\
      usleep(10);\
   }while(0);

   sendtcppacket (0, TH_SYN);
   sendtcppacket (1, TH_SYN | TH_ACK);
   sendtcppacket (2, TH_FIN | TH_ACK);
   sendtcppacket (3, TH_SYN | TH_FIN);
   sendtcppacket (4, TH_SYN | TH_XXX | TH_YYY);
   sendtcppacket (5, TH_SYN); /* this extra syn is used for hamming distances
                               * an even number of responses are needed */
   timeout = time (NULL) + timedelay;
   n = i = 0;

   do {
      if (get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
         if (dst_port == tcp.th_sport) {
            if (ntohs(tcp.th_dport) - start < PACKETS) {
               n = ntohs(tcp.th_dport) - start;
               if ((n < 0) || (n >= PACKETS)) continue;
               if (pak[n].set) continue;
               if (!n && !ntohs(tcp.th_win)) return(NULL);
               pak[n].seq = tcp.th_seq != 0;
               if (pak[n].seq) {
                  seq_mask ^= tcp.th_seq;
                  last_seq = tcp.th_seq;
                  i++;
               }
               pak[n].ack = ntohl (tcp.th_ack);
               pak[n].ackset = tcp.th_ack != 0;
               if (pak[n].ackset && myseq[n] > pak[n].ack)
                  pak[n].ackset = 2;
               if (pak[n].ackset)
                  pak[n].ack = (myseq[n] < pak[n].ack) ?
                               (pak[n].ack - myseq[n]) :
                               (myseq[n] - pak[n].ack);
               if (pak[n].ack > ACK_RAND) {
                  pak[n].ackset = 3;
                  pak[n].ack = 0;
               }
               pak[n].win = ntohs (tcp.th_win);
               pak[n].flag = tcp.th_flags;
	       pak[n].set = 1;
               pak[n].urg = tcp.th_urp != 0;
            }
         }
      }
      usleep (10);
   } while(timeout > time(NULL) && n != PACKETS - 1);
  
   /* calculate hamming distance */
   if(i % 2) seq_mask ^= last_seq;
   ham = hamming(seq_mask);
   if(ham > 8) ham = 16 - ham;

   /* parse fingerprints */
   if((fs_cfg = fopen(OSCFGFILE, "r"))) {
      struct fingerprint parse;
      int flag1 = 0, found = 0, total = 0;
      char *token;

      while(fgets(line, 80, fs_cfg)) {
         if(*line == '#') continue; /* COMMENT LINE */
         if(*line == '\n' || !strchr(line, '\n')) { /* NEW LINE */
            if(flag1 && (found == total)) {
               fclose(fs_cfg);
               *osname = (char)ham + '0';
               return osname;
            } else continue;
         }
         
         if(*line == '*') { /* OS FINGERPRINT HEADER */
            if(strlen(line) > 3) {
               strcpy(osname, &line[1]);
               osname[strlen(osname) - 1] = '\0';
            }
            flag1 = 1;
            total = found = 0;
            continue;
         }

         token = strtok(line, " ");
         if(isdigit(*token)) n = atoi(token);
         else {
            flag1 = total = found = 0;
            continue;
         }

         if((token = strtok(NULL, " ")))
            parse.seq = atoi(token);

         parse.set = isdigit(*token);

	 if((token = strtok(NULL, " "))) {
            switch(*token) {
               case '+':
                  parse.ackset = 1;
		  parse.ack = atol(token + 1);
		  break;
               case '-':
	          parse.ackset = 2;
                  parse.ack = atol(token + 1);
		  break;
               case 'R':
                  parse.ackset = 3;
		  parse.ack = 0;
		  break;
               case 'N':
               default:
                  parse.ackset = parse.ack = 0;
		  break;
            }
          }

	  if((token = strtok(NULL, " ")))
             parse.win = strtol(token, NULL, 16);

	  if((token = strtok(NULL, " \n"))) {
             parse.flag = 0;
	     if(strchr(token, 'S')) parse.flag |= TH_SYN;
	     if(strchr(token, 'R')) parse.flag |= TH_RST;
	     if(strchr(token, 'A')) parse.flag |= TH_ACK;
	     if(strchr(token, 'F')) parse.flag |= TH_FIN;
	     if(strchr(token, 'P')) parse.flag |= TH_PUSH;
	     if(strchr(token, 'X')) parse.flag |= TH_XXX;
	     if(strchr(token, 'Y')) parse.flag |= TH_YYY;
	     if(strchr(token, 'U')) parse.urg = 1;
	     else parse.urg = 0;
          }

	  total++;

	  /* test for no response */
	  if(!pak[n].set) {
             if(parse.set) total = found = flag1 = 0;
             else found++;
             continue;
          }

	  if((parse.ackset == pak[n].ackset) &&
	     (parse.ack == pak[n].ack) &&
	     (parse.seq == pak[n].seq) &&
	     (parse.win == pak[n].win) &&
	     (parse.flag == pak[n].flag) &&
	     (parse.urg == pak[n].urg))
             found++;
          else total = found = flag1 = 0;
      }

      fseek(fs_cfg, 0L, SEEK_END);
      fclose(fs_cfg);
      return NULL;
   } else {
      puts("error: cannot open " OSCFGFILE);
      exit(EXIT_FAILURE);
   }
}

void show_print(FILE *fsp) {
   int i;
   
   fprintf(fsp, "\n\t\t* Unknown OS Fingerprint");
   for(i = 0; i < PACKETS - 1; i++) {
      if(pak[i].set) {
         fprintf(fsp, "\n\t\t%i %i ", i, pak[i].seq);
         switch(pak[i].ackset) {
            case 0: case 3:
               fputc(pak[i].ackset ? 'R' : 'N', fsp); break;
            case 1: case 2:
               fprintf(fsp, "%c%li", pak[i].ackset == 1 ? '+' : '-',
                  pak[i].ack);
               break;
         }
         fprintf(fsp, " %04x ", pak[i].win);
         if(pak[i].flag & TH_SYN ) fputc('S', fsp);
         if(pak[i].flag & TH_RST ) fputc('R', fsp);
         if(pak[i].flag & TH_ACK ) fputc('A', fsp);
         if(pak[i].flag & TH_FIN ) fputc('F', fsp);
         if(pak[i].flag & TH_PUSH) fputc('P', fsp);
         if(pak[i].flag & TH_XXX ) fputc('X', fsp);
         if(pak[i].flag & TH_YYY ) fputc('Y', fsp);
         if(pak[i].urg           ) fputc('U', fsp);
      } else fprintf(fsp, "\n\t\t%i - - - -", i);
   }
}
